#pragma once

#ifdef _HAS_EXCEPTIONS
#undef _HAS_EXCEPTIONS
#endif
#define _HAS_EXCEPTIONS 0

#include <vector>

#pragma warning(disable: 4996)

typedef unsigned int uint;
typedef unsigned long long ull;
const uint MAX_UINT = 0xFFFFFFFF;
#define null 0

#ifdef _WIN32
	#include <intrin.h>		// __rdtsc()
#else
	#include <stdint.h>
	extern "C" 
	{
		__inline__ unsigned long long __rdtsc() 
		{
			uint32_t lo, hi;
			__asm__ __volatile__ (
				"xorl %%eax,%%eax \n        cpuid" ::: "%rax", "%rbx", "%rcx", "%rdx");
			/* We cannot use "=A", since this would use %rax on x86_64 and return only the lower 32bits of the TSC */
			__asm__ __volatile__ ("rdtsc" : "=a" (lo), "=d" (hi));
			return (uint64_t)hi << 32 | lo;
		}
	}
#endif

struct Edge;

struct Field
{
	uint uIndex;
	uint* puEdges;		// coresponding edges
	uint uEdgesCount;
};

struct Edge
{
	uint uField[2];
};

typedef uint (*SelectIndividualCallback)(const uint* pGenerationConflicts, uint uGenerationSize, uint uExcept);	// return value must differ from uExcept
typedef void (*CrossoverIndividualsCallback)(const uint* pFirstInd, const uint* pSecondInd, uint uIndividualSize, uint* pNewFirstInd, uint* pNewSecondInd);
typedef void (*MutateIndividualCallback)(uint* pIndividual, uint uIndividualSize, uint uColorsCount);

class GraphColoring
{
	typedef std::vector<Edge> EdgeVector;
	typedef std::vector<Field> FieldVector;
	typedef std::vector<uint> UintVector;

public:
	GraphColoring(uint uIndividiuals,
					double dCrossingProbability,
					double dMutatationProbability,
					SelectIndividualCallback cbSelectCallback,
					CrossoverIndividualsCallback cbCrossoverCallback,
					MutateIndividualCallback cbMutationCallback,
					uint uColorsCount) 
		: m_uGenerationSize(uIndividiuals)
		, m_uActualGeneration(0)
		, m_uColorsCount(uColorsCount)
		, m_pBestGeneration(null)
		, m_pPrevGeneration(null)
		, m_pActualGeneration(null)
		, m_pBestGenerationConflicts(null)
		, m_pPrevGenerationConflicts(null)
		, m_pActualGenerationConflicts(null)
		, m_dCrossingProbability(dCrossingProbability)
		, m_dMutatationProbability(dMutatationProbability)
		, m_cbSelectIndividual(cbSelectCallback)
		, m_cbCrossoverIndividuals(cbCrossoverCallback)
		, m_cbMutateIndividual(cbMutationCallback)
	{
	}

	bool LoadGraph(const char* szGraphFile);
	void ReleaseGraph();

	void ResetGraph(uint uColorsCount);

	void InitPopulation();

	void IterateGeneration();

	uint GetBestGeneration() { return m_uBestIndividual[0]; }
	uint GetBestIndividual() { return m_uBestIndividual[1]; }
	uint GetBestConflictsCount() { return *GetIndividualConflictsOffset(m_pBestGenerationConflicts, m_uBestIndividual[1]); }
	uint GetBestFieldColor(uint uIndex) { return GetIndividualOffset(m_pBestGeneration, m_uBestIndividual[1])[uIndex]; }

	uint GetFieldsCount() { return m_aFields.size(); }
	uint GetColorsCount() { return m_uColorsCount; }

private:
	uint* GetIndividualOffset(uint* pGenerationOffset, uint uIndividualIndex) { return pGenerationOffset + uIndividualIndex * m_aFields.size(); }

	uint* GetIndividualConflictsOffset(uint* pGenerationConflictsOffset, uint uIndividualIndex) { return pGenerationConflictsOffset + uIndividualIndex; }

	void SwapGenerations();

	uint EvaluateConflicts(uint* uIndividualOffset);

	void EvaluateGeneration(uint* pGeneration, uint* pGenerationConflicts, uint uGeneration);

	static bool CheckProbability(double dCoeff);
	
private:
	FieldVector m_aFields;
	EdgeVector m_aEdges;	

	uint m_uActualGeneration;
	uint m_uColorsCount;
	double m_dCrossingProbability;		// probability [0.0; 1.0]
	double m_dMutatationProbability;	// probability [0.0; 1.0]

	SelectIndividualCallback m_cbSelectIndividual;
	CrossoverIndividualsCallback m_cbCrossoverIndividuals;
	MutateIndividualCallback m_cbMutateIndividual;

	uint* m_pPrevGeneration;
	uint* m_pActualGeneration;
	uint* m_pBestGeneration;
	uint* m_pPrevGenerationConflicts;
	uint* m_pActualGenerationConflicts;
	uint* m_pBestGenerationConflicts;

	uint* m_aHistory;				// Three generations in single memory (only colors, rest need to be calculated)
	uint* m_aHistoryConflicts;		// Number of conflicts in three generations (best, previous, actual)
	uint m_uGenerationSize;

	uint m_uBestIndividual[2];
};
